#!/usr/bin/perl -w
#!/usr/local/bin/perl -w
#!/bin/perl -w

## Author: Jeong-Hyeon Choi
## Date: 2010

## Description

## Usage:
## CHECK usage with option  -h

## $Id: perl_template,v 1.2 2007-09-21 22:51:23 jeochoi Exp $


use Getopt::Long qw(:config no_ignore_case);

my ($helpFlag, $inFile, $outFile, $verbose, $qual);

GetOptions(
	"h|?|help"		=> \$helpFlag,	
	"input=s"		=> \$inFile,		## input file
	"output=s"		=> \$outFile,		## output file
	"verbose+"		=> \$verbose,		## verbose output
	"qual"			=> \$qual,
) || die "\n";

checkOptions();

load($inFile);


#-------------------------------------------------------------------------------

sub load
{
	my ($fileName) = @_;

	print "Loading ", $fileName || 'STDIN', " ...\n" if $inFile && $outFile;
	my $in = openInput($fileName);
	my $out = openOutput($outFile);
	my @old;

	while (<$in>)
	{
		next if /^#/ || /^\s*$/;
		my @a = split (/\t/);

		if (@old)
		{
			if ($old[0] eq $a[0] && $old[3] == $a[3])
			{
				my ($m1, $u1, $m2, $u2);
				if (!$qual)
				{
					die "Wrong format in $_" if $old[8] !~ /mread=(\d+);uread=(\d+)/; ($m1, $u1) = ($1, $2);
					die "Wrong format in $_" if $a  [8] !~ /mread=(\d+);uread=(\d+)/; ($m2, $u2) = ($m1+$1, $u1+$2);
					$old[8] =~ s/mread=(\d+);uread=(\d+)/mread=$m2;uread=$u2/;
				}
				else
				{
					die "Wrong format in $_" if $old[8] !~ /mqual=(\d+);uqual=(\d+)/; ($m1, $u1) = ($1, $2);
					die "Wrong format in $_" if $a  [8] !~ /mqual=(\d+);uqual=(\d+)/; ($m2, $u2) = ($m1+$1, $u1+$2);
					$old[8] =~ s/mqual=(\d+);uqual=(\d+)/mqual=$m2;uqual=$u2/;
				}

				$old[5] = sprintf("%.3f", $m2+$u2 == 0 ? -1 : $m2/($m2+$u2));
				$old[6] = '.';
			}
			else
			{
				print $out join("\t", @old);
				@old = @a;
			}
		}
		else
		{
			@old = @a;
		}
	}

	print $out join("\t", @old) if @old;
	close($in) if defined $fileName;
}


#-------------------------------------------------------------------------------

sub openInput
{
	my ($fileName) = @_;

	return *STDIN unless defined $fileName;

	my ($fd);
	open($fd, $fileName =~ /.gz(ip)?$/ ? "zcat $fileName |" : $fileName =~ /.bz(ip)?2$/ ? "bzcat $fileName |" : $fileName) || die("Open error: $fileName");
	return $fd;
}

sub openOutput
{
	my ($fileName) = @_;

	return *STDOUT unless defined $fileName;

	my ($fd);
	open($fd, $fileName =~ /.gz$/ ? "| gzip -c > $fileName" : $fileName =~ /.bz(ip)?2$/ ? "| bzip2 -z -c > $fileName" : ">$fileName") || die("Open error: $fileName");
	return $fd;
}

sub checkOptions
{
	$inFile  = shift(@ARGV) if !defined $inFile  && @ARGV > 0;
	$outFile = shift(@ARGV) if !defined $outFile && @ARGV > 0;
	die "The output file name is the same as the input file name\n" if defined $inFile && defined $outFile && $inFile eq $outFile;

	if ($helpFlag)
	{
		die("Arguments: [[-i] in_file] [[-o] out_file] [-v] [-q]\n"
		  . "\t\n"
		  );
	}
}
