#!/usr/bin/perl -w
#!/usr/local/bin/perl -w
#!/bin/perl -w

## Author: Jeong-Hyeon Choi
## Date: 2010

## Description

## Usage:
## CHECK usage with option  -h

## $Id: perl_template,v 1.2 2007-09-21 22:51:23 jeochoi Exp $

use Getopt::Long qw(:config no_ignore_case);

my ($helpFlag, $inFile, $outPrefix, $verbose, $quiet, $minRead, $track, $name, $gzip);

GetOptions(
	"h|?|help"		=> \$helpFlag,	
	"input=s"		=> \$inFile,		## input file
	"output=s"		=> \$outPrefix,		## output file
	"verbose+"		=> \$verbose,		## verbose output
	"quiet"			=> \$quiet,
	"min=i"			=> \$minRead,
	"name=s"			=> \$name,
	"track=s"		=> \$track,
	"gzip"			=> \$gzip,
) || die "\n";

checkOptions();

load($inFile);


#-------------------------------------------------------------------------------

sub load
{
	my ($fileName) = @_;

	print "Loading ", $fileName || 'STDIN', " ...\n" if !$quiet;
	my $in = openInput($fileName);
	my %outs;
	my $trackLine = '';

	if ($name)
	{
		$trackLine = "track name=$name itemRgb=On";
		$trackLine .= $track if $track;
		$trackLine .= "\n";
	}

	while (<$in>)
	{
		next if /^#/ || /^\s*$/;
		my @a = split(/[\t\r\n]/);

		my $out;
		my $type = lc ($a[2]);
		if (exists $outs{$type}) { $out = $outs{$type}; }
		else                     { $out = $outs{$type} = openOutput("$outPrefix.$type.bed"); print $out $trackLine; }

		die "Wrong format in $_" if $a[8] !~ /mread=(\d+);uread=(\d+)/;
		print $out join("\t", $a[0], $a[3]-1, $a[4], $1+$2, $a[5], $a[6], $1), "\n" if $1+$2 >= $minRead;
	}

	close($in) if defined $fileName;
}

#-------------------------------------------------------------------------------

sub openInput
{
	my ($fileName) = @_;

	return *STDIN unless defined $fileName;

	my ($fd);
	open($fd, $fileName =~ /.gz(ip)?$/ ? "zcat $fileName |" : $fileName =~ /.bz(ip)?2$/ ? "bzcat $fileName |" : $fileName) || die("Open error: $fileName");
	return $fd;
}

sub openOutput
{
	my ($fileName) = @_;

	return *STDOUT unless defined $fileName;

	my ($fd);
	open($fd, $fileName =~ /.gz$/ ? "| gzip -c > $fileName" : $gzip ? "| gzip -c > $fileName.gz" : $fileName =~ /.bz(ip)?2$/ ? "| bzip2 -z -c > $fileName" : ">$fileName") || die("Open error: $fileName");
	return $fd;
}

sub checkOptions
{
	$inFile  = shift(@ARGV) if !defined $inFile  && @ARGV > 0;
	$outPrefix = shift(@ARGV) if !defined $outPrefix && @ARGV > 0;

	if ($helpFlag || !$outPrefix)
	{
		die("Arguments: [-n name] [-t track] [[-i] in_file] [-o] out_prefix [-v] [-q]\n"
		  . "\t\n"
		  );
	}

	$minRead = 0 if !$minRead;
}
