#!/usr/bin/perl

use strict;
use Getopt::Long;
use Data::Dumper;

my $qual;
my $minQual;
my $out;

my %IUPAC=(
			"A"=>"A",
			"T"=>"T",
			"C"=>"C",
			"G"=>"G",
			"M"=>"AC",
			"R"=>"AG",
			"Y"=>"CT",
			"W"=>"AT",
			"S"=>"CG",
			"K"=>"GT",
			"B"=>"CGT",
			"D"=>"AGT",
			"H"=>"ACT",
			"V"=>"ACG",
			"N"=>"ATCG"
			
			);
			
sub checkBasFilter{
	my ($iupacStr)=@_;
	
	my @iupacS=split(//,$iupacStr);
	foreach (@iupacS){
		if(! exists $IUPAC{$_}){
			return $_;
		}
	}
	
	return -1;
}			

sub basFilter{
	my ($seq,$iupacStr)=@_;
	my @seqs=split(//,uc($seq));
	my @iupacS=split(//,$iupacStr);
	
	my $length=0;
	if( length($seq)<length($iupacStr) ){
		$length=length($seq);
	}else {
		$length=length($iupacStr);
	}
	
	for(my $i=0; $i<$length; ++$i){
		#print $seqs[$i]." ".$IUPAC{$iupacS[$i]}."\n";
		if(index($IUPAC{$iupacS[$i]},$seqs[$i])<0){
			return 0;
		}
	}
	#print '1'."\n";
	return 1;
	
}
sub DNAreverse {
	my %trans=("A"=>"T",
	    "T"=>"A",
	    "G"=>"C",
	    "C"=>"G",
	    "N"=>"N"
	);
	my $seq=uc(shift);
 	my @char=split(//,$seq );
	my @reverse=reverse(@char);
	
	my $re="";

	foreach my $element (@reverse){
		if(exists $trans{$element}){
		 	$re=$re.$trans{$element};
		}else {
			#print $element."\n";
		}
	}
	return $re;
}

sub getTag{
	my($line_array,$start)=@_;
	
	my %result;
	
	for(my $i=$start ; $i<scalar(@$line_array) ; $i++ ){
		my $tag=$$line_array[$i];
		my @splits=split(/:/,$tag);
		
		#print Dumper(\@splits);
		
		if(scalar(@splits) == 3){
			
			$result{$splits[0]}=$splits[2];
		}
		
		
	}
	
	return %result;
	
}

sub getStand{
	if(scalar(@_)<3){
		return ;
	}
	
	my ($result_hash,$refseqMeth_hash,$index_hash,$REFFILE,$readSeq_Fun)=@_;
	my $seqname="";
	my $sequence_ref="";
	
	foreach  (keys %$refseqMeth_hash){
			$seqname=$_;
			$sequence_ref=&$readSeq_Fun($REFFILE,$index_hash,$seqname);
		
			my $position_hash=$$refseqMeth_hash{$seqname};
			my @positions=keys %$position_hash;
		
			my ($chrome,$start,$end)=parseSeqname($seqname);
			
			#print "$seqname .. $sequence \n";
		
			foreach my $position (@positions) {
				my ($methylatedN,$downstream)=getMdownstream($sequence_ref,$position-$start+1);
				if(exists $$result_hash{$chrome}{$position}){
					$$result_hash{$chrome}{$position}[3]=$methylatedN;
					$$result_hash{$chrome}{$position}[4]=$downstream;
				
				}
			
			}
		
		
	}
	
	#my $i=0;

	#while(my $line=<FASTA>){
	#	chomp $line;
	
		#print $line."\n";
	#	if(index($line,">")>-1){
	#		if(length($seqname)>1){
				
				#printn "$seqname";
				#printn "$sequence";
	#			++$i;
				
	#			fillStandFollow($seqname,$sequence,$result_hash,$refseqMeth_hash);
				
	#		}
	#		$seqname=$line;
	#		$seqname=~s/\>//;
		
	#		$sequence="";
	#		next;		
	#	}
	#	$line=~s/\s//;
	#	$sequence=$sequence.$line;
	#}
	#	fillStandFollow($seqname,$sequence,$result_hash,$refseqMeth_hash);
				


}

sub parseSeqname{
	my $seqname=$_[0];
	
	my ($chrom,$start,$end)=split(/[:,\.]+/,$seqname);
	$start=1  unless (defined  $start); 
	return ($chrom,$start,$end);
}

sub fillStandFollow{
	if(scalar(@_)<4){
		return ;
	}
	my ($seqname,$sequence,$result_hash,$refseqMeth_hash)=@_;
	
	
	
	if(exists $$refseqMeth_hash{$seqname}){
		my $position_hash=$$refseqMeth_hash{$seqname};
		my @positions=keys %$position_hash;
		
		my ($chrome,$start,$end)=parseSeqname($seqname);
		
		
		foreach my $position (@positions) {
			my ($methylatedN,$downstream)=getMdownstream($sequence,$position-$start+1);
			if(exists $$result_hash{$chrome}{$position}){
				$$result_hash{$chrome}{$position}[3]=$methylatedN;
				$$result_hash{$chrome}{$position}[4]=$downstream;
				
			}
			
		}
		
		
	}else{
		return ;
	}
	
}

sub getMdownstream{
		my ($refseq_ref,$position)=@_;
			
		
			my $methylatedN=uc(substr($$refseq_ref,$position-1,1));
			my $downstream;
				      if ( $methylatedN eq "C"){
				 		$downstream=substr($$refseq_ref,$position,2);
				      }
				 	  if ($methylatedN eq "G") {
				 	  	my $rUpdown=uc(substr($$refseq_ref,$position-3,2));
				 	  	$downstream=DNAreverse($rUpdown);
				 	  	
				 	  }
				
			
		
		
		return ($methylatedN,$downstream);
}

sub getPgname{
	my ($line)=@_;
	my @temp=split(/\t/,$line);
	
	shift @temp;
	
	foreach (@temp){
		my @headtag=split(/:/,$_);
		if( $headtag[0] eq "ID"){
			return $headtag[1];
		}
	}
	
	return "";
}

sub readFilter{
	my ($filter,$filter_hash)=@_;
	
	open(FILTER,$filter) || die "can't open file $filter.";
	
	while(<FILTER>){
		my ($chrome,$pos)=split(/\t/,$_);
		$$filter_hash{$chrome}{$pos}=0;
	}
	
}

sub output{
	my ($result_hash,$chrome, $position,$length,$pattern,$program)=@_;
	
	my	$offset=$length;
	my	$methy=$$result_hash{$chrome}{$position}[0];
	my	$unmethy=$$result_hash{$chrome}{$position}[1];
	my	$reads=$$result_hash{$chrome}{$position}[2];
		
	my	$ncbase=uc($$result_hash{$chrome}{$position}[3]);
		
	my  $strand= $ncbase eq "C" ? "+" : "-" ;
	my	$updown= uc( $$result_hash{$chrome}{$position}[4] );
	
	my $mQual=$$result_hash{$chrome}{$position}[5];
	my $uQual=$$result_hash{$chrome}{$position}[6];
	
	if(length($updown)>$length){
		$updown=substr($updown,0,$length);
		
	}else {
		$offset=length($updown);
	}
	
	my $position2=$position;
	
	#print $position."\n";
	if($position =~  /^\d+$/){
		if($strand eq "+"){
			$position2=$position+$offset;
		}else {
			$position2=$position;
			$position=$position-$offset;
		}
	#	print $position."\n";
	}else {
		$position2=$position
	}
	
		$pattern= $pattern ne ""?$pattern:"C$updown";
		
			my 	$score=0;
			if($qual){
				if($mQual+$uQual>0){
					$score=$mQual/($mQual+$uQual);
				}else {
					$score=-1;
				}
			}else{
				$score=$methy/($methy+$unmethy);
			}
		
			$score=sprintf("%.3f",$score);
	
			print $out $chrome."\t";
 			print $out "$program"."\t";
 			print $out "$pattern"."\t";
 			print $out $position."\t";
 			print $out $position2."\t";
 			print $out $score."\t";
 			print $out "$strand\t.\t";
 			print $out "mread=$methy;uread=$unmethy";
 			if($qual){
 				print $out ";mqual=$mQual;uqual=$uQual";
 			}
 			print $out ";site=C$updown";
 			print $out "\n"; 
}

sub getQual{
	
	my ($quality, $pos)=@_;
	
	if(length($quality)==1){
		return 0;
	}
	my $pQual=substr $quality,$pos-1,1;
	return ord($pQual)-33;
}

sub printHelp{
	print " msam2gff.pl convert sam format to gff format.\n";
	print " msam2gff.pl -sam SamFile -o OutputFile -seq referenceFile.fasta ";
	print "[-qual] [-mq min_base_qual] [-filter filterFile] -f IUPACs[:IUPACs] [-p ProgramName]\n";
	#print "-n how many sites will be loaded in memory";
}


#sub outputTempFile{ 
#	my ($result_hash,$posfilter_hash,$tempfile)=@_;
#	
#	open(TEMP,">$tempfile") || die "can't open tempfile $tempfile";
#			outputFile($result_hash,$posfilter_hash,*TEMP,1);
#			
#			 close TEMP;
#}

sub isOverlap{
	my ($chrom,$start,$preChrom,$preEnd)=@_;
	
	if( $chrom eq $preChrom && $preEnd>=$start){
		 return 1;
	}else {
		return 0;
	}
}
sub makeIndex{
my  $file=shift;
my $index_hash=shift;

my $preoffset=0;
my $seqname="";
my $seqstart=0;
my $seqlength=0;

	while(<$file>){
		chomp $_;
		if(/^>/){
			if(length $seqname > 1){
				$$index_hash{$seqname}=[$seqstart,$preoffset,$seqlength];
			}
			$seqname = substr($_,1);
			$seqstart=tell($file);
			$seqlength=0;
		}
		$seqlength+=length($_);
		$preoffset=tell($file);
	}
	if(! exists $$index_hash{$seqname}){
		$$index_hash{$seqname}=[$seqstart,$preoffset,$seqlength];
	}
}

sub getSequenceLength{
	my $index_hash=shift;
	my $seqname=shift;
	
	if(exists $$index_hash{$seqname}){
		return $$index_hash{$seqname}->[2];
	}
	return 0;
}

sub creadsequence{
	my $cSeq="";
	my $cSeqname="";
	my $rs=sub{
		my ($file, $index_hash,$seqname)=@_;
		if(!defined $seqname){
			return "";
		}
		if($cSeqname eq $seqname){
			return \$cSeq;
		}else {
			$cSeqname=$seqname;
			$cSeq=readsequence($file,$index_hash,$seqname);
			return \$cSeq;
		}
		
	};
	return $rs;
}

sub readsequence{
	my $file=shift;
	my $index_hash=shift;
	my $seqname=shift;
	
	if(!exists $$index_hash{$seqname}){
		return "";
	}
	
	my $start=$$index_hash{$seqname}->[0];
	my $end  =$$index_hash{$seqname}->[1];
	
	seek($file,$start,0);

	my $sequence="";

	read ($file,$sequence ,$end-$start);
	#print "readfile\n";
	$sequence=~ s/\n//g;
	$sequence=~ s/(.)/\U$1\E/g;
	return $sequence;
}



my $samfile="";
my $output="";
my $seqfile="";

my $filter="";
my $filter_bases;
my %filter_na;
my $inputProgram="";



 Getopt::Long::GetOptions(              
      'sam=s'    => \$samfile,            
       'o=s'    => \$output,           
       "seq=s"=>\$seqfile,          
       'filter=s'    => \$filter,           
       'n=s' => \$filter_bases,
       'p=s' => \$inputProgram,
       'mq' => \$minQual,
       'qual' => \$qual); 
 

my %posfilter;

if( ($samfile eq "") || ($seqfile eq "") ){
	printHelp();
	exit(1);
}

if(length($filter)>2){
	readFilter($filter,\%posfilter);
}

$filter_bases = 'G:H' if !$filter_bases;
if($filter_bases ne ""){
	my @temp=split(/:/,uc($filter_bases));
	
	my $check=checkBasFilter(join("",@temp));
	
	if($check!=-1){
		die "$filter_bases: $check not in IUPAC.";
	}
	
	foreach (@temp){
		$filter_na{uc($_)}=0;
	}
}

open(FILE,$samfile =~ /\.bam$/ ? "samtools view -h $samfile |" : $samfile)|| die ("Could not open file ".$samfile);


$out = openOutput($output);;


my %result;
my %refseqMeth;
my $startSeq=0;
my $program="";



my $preChrom="";
my $preMatchEnd="";

open(REF, "$seqfile" )|| die  "can't open file $seqfile. $@";
my %index;
makeIndex(*REF,\%index);

 my  $readseq_Fun= creadsequence();

while(<FILE>){
	
	chomp $_;
	my $fistchar=substr($_,0,1);
	if(index($_,"\@PG")>-1){
			$program=getPgname($_);
			#print "$program\n";
			
			#if user input programme name, just use it.
		if($inputProgram ne ""){
			$program=$inputProgram;
		}
			
			
			next;
		}
		
	if (!($fistchar eq "\@")) {
		$startSeq=1;
		#print "$_  ".index($_,"\@PG")."\n";
		
	
	}

	if($startSeq>0){
		#print $_."\n";
		my @line=split(/\t/,$_);
		my $refseq=$line[2];
		my ($chrome,$start,$end)=parseSeqname($line[2]);
		my $offset=$line[3];
		my $pos=$line[3];
		my $seq=$line[9];
		my $quality=$line[10];
		my $matchStart=$start+$offset;
		my $matchEnd=$matchStart + length($seq);
		
		if(!isOverlap($chrome,$matchStart,$preChrom,$preMatchEnd)){
			getStand(\%result,\%refseqMeth,\%index,*REF,$readseq_Fun);
			outputFile(\%result,\%posfilter,$filter,$filter_bases,\%filter_na,$program);
			 undef %refseqMeth;
			 undef %result;
		}
		
		my %tag=getTag(\@line,11);
		#print Dumper(\%tag);
		my @meths  = exists $tag{"YM"} ? split(/,/,$tag{"YM"}) : ();
		my @umeths = exists $tag{"YU"} ? split(/,/,$tag{"YU"}) : ();
		
		for(my $i=0;$i<scalar(@meths);++$i){
			my $meth=$meths[$i];
			my $meth_Qual=getQual($quality,$meth);
			next if $meth_Qual < $minQual;
			my $meth_position=$start+$offset+$meth-2;
			
			if(exists $result{$chrome}{$meth_position}){
				 $result{$chrome}{$meth_position}[0]= $result{$chrome}{$meth_position}[0]+1;
				# print "hit\n"
				 $result{$chrome}{$meth_position}[5]= $result{$chrome}{$meth_position}[5]+$meth_Qual;
		
			}else {
				$result{$chrome}{$meth_position}=[(1 ,0 ,$meth_position,"","",$meth_Qual,0)];
				
			}
			
			if(exists $refseqMeth{$refseq}){
				my $position_hash=$refseqMeth{$refseq};
				if (!(exists $$position_hash{$meth_position})){
					$$position_hash{$meth_position}=0;
				}
			 }else {
			 	my %position=($meth_position=>0);
			 	$refseqMeth{$refseq}=\%position;
			 }
		}
		
		for(my $i=0;$i<scalar(@umeths);++$i){
			my $umeth=$umeths[$i];
			my $umeth_Qual=getQual($quality,$umeth);
			next if $umeth_Qual < $minQual;
			my $umeth_position=$start+$offset+$umeth-2;

			if(exists $result{$chrome}{$umeth_position}){
				 $result{$chrome}{$umeth_position}[1]= $result{$chrome}{$umeth_position}[1]+1;
				 $result{$chrome}{$umeth_position}[6]= $result{$chrome}{$umeth_position}[6]+$umeth_Qual;
				
			}else {
				$result{$chrome}{$umeth_position}=[(0 ,1 ,$umeth_position,"","",0,$umeth_Qual)];
			}
			
			if(exists $refseqMeth{$refseq}){
				my $position_hash=$refseqMeth{$refseq};
				if (!(exists $$position_hash{$umeth_position})){
					$$position_hash{$umeth_position}=0;
				}
			 }else {
			 	my %position=($umeth_position=>0);
			 	$refseqMeth{$refseq}=\%position;
			 }
			
			
		}
		
		#if(scalar(keys %result) > $limit){
		#	 print scalar(%result)."\n";
		#	 getStand(\%result,\%refseqMeth,$seqfile);
		#	 outputTempFile(\%result,\%posfilter,$output.".".$tempFileNumber);
		#	 $tempFileNumber++;
		#	 undef %result;
		#	 }
		
		$preMatchEnd=$matchEnd if ($preChrom && $preChrom ne $chrome) || !$preMatchEnd || $preMatchEnd < $matchEnd;
		$preChrom=$chrome;
		
	}	
}

close FILE;

getStand(\%result,\%refseqMeth,\%index,*REF,$readseq_Fun);
outputFile(\%result,\%posfilter,$filter,$filter_bases,\%filter_na,$program);

close REF;

#print "\n";
#print scalar(%result);
#print "\n", scalar(keys %result), "\n";



sub outputFile{
	my ($result_hash,$posfilter_hash,$filter,$filter_bases,$filter_na_hash,$program)=@_;
	
	foreach my $chr (sort keys %$result_hash){
		foreach my $pos (sort {$a<=>$b} keys %{$result_hash->{$chr}}){
		if(length($filter)>2 && scalar(keys %$posfilter_hash)>1){
			if(!(exists $$posfilter_hash{$chr}{$pos})){
				next;
			}
		}
	
			my	$updown= uc( $$result_hash{$chr}{$pos}[4] );
		
			if($filter_bases ne "" ){
				foreach (keys %$filter_na_hash){
					if(basFilter($updown,$_)==1){
						output($result_hash,$chr, $pos,length($_),"C".$_,$program);
					}
				}	
			}else {
	 			output($result_hash,$chr, $pos,length($updown),"",$program);
			}
		}
	}
}

sub openOutput
{
	my ($fileName) = @_;

	return *STDOUT unless $fileName;

	my ($fd);
	open($fd, $fileName =~ /.gz$/ ? "| gzip -c > $fileName" : $fileName =~ /.bz(ip)?2$/ ? "| bzip2 -z -c > $fileName" : ">$fileName") || die("Open error: $fileName");
	return $fd;
}
